/*
  ogmmerge -- utility for splicing together ogg bitstreams
      from component media subtypes

  avi.c
  ogmmerge AVI (video only) module

  Written by Moritz Bunkus <moritz@bunkus.org>

  Copyright 2001 Ralph Giles <Ralph_Giles@telus.net>

  Distributed under the GPL
  see the file COPYING for details
  or visit http://www.gnu.org/copyleft/gpl.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <ogg/ogg.h>
#include <avilib.h>

#include "ogmmerge.h"
#include "ogmstreams.h"
#include "util.h"
#include "avi.h"

typedef struct {
  ogg_stream_state  os;
  int               serialno;
  u_int64_t         packetno;
  u_int64_t         old_granulepos;
  double            fps;
  unsigned char    *chunk;
  avi_t            *avi;
  q_t              *q;
} avi_state_t;

u_int64_t avi_make_timestamp(ogmmerge_state_t *omst, ogg_int64_t granulepos) {
  u_int64_t    stamp;
  avi_state_t *state = (avi_state_t *)omst->private;

/*  stamp = (double)state->packetno * (double)1000000 /
          (double)state->fps;*/

  stamp = (double)granulepos * (double)1000000 /
          (double)state->fps;

  state->old_granulepos = granulepos;

  return stamp;
}

/* allocates and initializes local storage for a particular
 * substream conversion.
 *
 * returns 1 on success, 0 on failure, 2 on success with the indication
 * that the module handles file access itself
 */
int avi_state_init(ogmmerge_state_t *state, int serialno, char *atracks,
                   char *vtracks)
{
  avi_state_t      *local;
  ogg_packet        op;
  
  if (state == NULL)
    return 0;
  
  local = (avi_state_t *)mymalloc(sizeof(*local));
  if (local == NULL)
    return 0;
  
  ogg_stream_init(&local->os, serialno);
  local->serialno = serialno;
  local->packetno = 0;
  local->chunk = mymalloc(1600 * 1200 * 3);
  if (local->chunk == NULL)
    return 0;
  local->old_granulepos = 0;
  state->private = local;
  if ((local->avi = AVI_open_input_file(state->name, 1)) == NULL)
    return 0;
  local->fps = AVI_frame_rate(local->avi);
  local->q = q_init(state, &local->os, avi_make_timestamp);

  if ((atracks != NULL) && (vtracks != NULL))
    fprintf(stderr, "Warning: -a/-A/-d/-D need still to be implemented " \
            "for AVI files. Processing video only.\n");
  
  return 2;
}

/* closes and deallocates any resources associated
 * with a substream conversion
 *
 * returns 0 on success
 */
int avi_state_close(ogmmerge_state_t *state)
{
  avi_state_t *local;
  
  if (state == NULL)
    return 0;
  local = (avi_state_t *)state->private;
  if (local != NULL) {
    ogg_stream_destroy(&local->os);
    state->private = NULL;
    AVI_close(local->avi);
    q_done(local->q);
    myfree(local->chunk);
    myfree(local);
  }
  return 0;
}

/* Normally accepts a buffer of input data 
 * the buffer must be completely consumed before returning
 * 
 * This module handles the file reading itself. buffer will be NULL,
 * size 0.
 *
 * returns 0 on success, a negative error from ogmmerge.h on failure
 */
int avi_data_in(ogmmerge_state_t *state, char *buffer, unsigned long size)
{
  avi_state_t   *avistate;
  avi_t         *avi;
  ogg_packet    *op;
  ogg_page      *page;
  stream_header  sh;
  char          *buf;

  if (state == NULL)
    return EOTHER;

  avistate = (avi_state_t*)state->private;
  avi = avistate->avi;
  if (avi == NULL)
    return EOTHER;
   
  /* are we looking for the signature? */
  if (avistate->packetno == 0) {
    char *codec = AVI_video_compressor(avi);
    long k, fsize = 0;
    memset(&sh, 0, sizeof(sh));
    strcpy(sh.streamtype, "video");
    memcpy(sh.subtype, codec, 4);
    sh.size = sizeof(sh);
    sh.time_unit = (long) (10000000.0 / AVI_frame_rate(avi));
    sh.samples_per_unit = 1;
    sh.default_len = 1;
    for (k = 0; k < AVI_video_frames(avi); k++)
      if (AVI_frame_size(avi, k) > fsize)
        fsize = AVI_frame_size(avi, k);
    sh.buffersize = fsize;
    sh.bits_per_sample = 24; // fixme!
    sh.sh.video.width = AVI_video_width(avi);
    sh.sh.video.height = AVI_video_height(avi);
  
    op = (ogg_packet *)mymalloc(sizeof(*op));
    if (op == NULL)
      die("malloc");
    buf = mymalloc(sizeof(sh) + 1);
    if (buf == NULL)
      die("malloc");
    *buf = PACKET_TYPE_HEADER;
    memcpy(buf + 1, &sh, sizeof(sh));
    op->packet = buf;
    op->bytes = 1 + sh.size;
    op->b_o_s = 1;
    op->e_o_s = 0;
    op->packetno = 0;
    op->granulepos = 0;
    /* submit it */
    ogg_stream_packetin(&avistate->os, op);
    avistate->packetno++;
    q_flush_pages(avistate->q);
  }

  if (avistate->packetno <= AVI_video_frames(avi)) {
    int nread, key;
    nread = AVI_read_frame(avi, avistate->chunk + 1, &key);
    if (nread < 0)
      return EOTHER;
    
    op = (ogg_packet *)mymalloc(sizeof(*op));
    if (op == NULL)
      die("malloc");
    avistate->chunk[0] = key ? PACKET_IS_SYNCPOINT : 0;
    op->bytes = nread + 1;
    op->packet = avistate->chunk;
    op->b_o_s = 0;
    op->e_o_s = (avistate->packetno == AVI_video_frames(avi) ? 1 : 0);
    op->granulepos = avistate->packetno;
    op->packetno = avistate->packetno++;
    ogg_stream_packetin(&avistate->os, op);
    q_queue_pages(avistate->q);
    return EMOREDATA;
  }

  return 0;  /* success */
}

/* returns an completed ogmmerge_page created from the data previously
 * passed through mng_data_in()
 *
 * returns NULL if no page is available
 */
ogmmerge_page_t *avi_page_out(ogmmerge_state_t *state) {
  avi_state_t *local = (avi_state_t *)state->private;
  
  return q_page_out(local->q);
}

