#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <ogg/ogg.h>
#include <vorbis/codec.h>

#include "ogmstreams.h"

#define MAX_AUDIO_TRACKS 20
#define MAX_VIDEO_TRACKS 20
#define BLOCK_SIZE 4096

struct audio_t {
  int bitrate;
  int channels;
  int rate;
};

struct demux_t {
  int              serial;
  int              fd;
  int              vorbis;
  ogg_stream_state state;
};

struct audio_t audio[MAX_AUDIO_TRACKS];
int            xaudio[MAX_AUDIO_TRACKS];
int            xvideo[MAX_VIDEO_TRACKS];

char *basename;

int verbose = 0;

void usage(char *fname) {
  fprintf(stdout, "Usage: %s [-v|-o outname|-a track|-nv] -i inname\n",
          fname);
  fprintf(stdout, "       -i inname   Use 'inname' as the source.\n");
  fprintf(stdout, "       -o outname  Use 'outname' as the base for " \
          "destination file names.\n");
  fprintf(stdout, "                   '-v1', '-v2', '-a1', '-a2'... will be " \
          "appended to this name.\n");
  fprintf(stdout, "                   Default: use 'inname'.\n");
  fprintf(stdout, "       -a track    Extract specified audio track. Can be " \
          "used more than once.\n");
  fprintf(stdout, "                   Default: extract all tracks.\n");
  fprintf(stdout, "       -d track    Extract specified video track. Can be " \
          "used more than once.\n");
  fprintf(stdout, "       -na         Don't extract any audio track.\n");
  fprintf(stdout, "       -nv         Don't extract any video track.\n");
  fprintf(stdout, "                   Default: extract all tracks.\n");
  fprintf(stdout, "       -v          Increase verbosity.\n");
  fprintf(stdout, "       -h          Show this help.\n");
}

void process_ogm(int fdin)
{
  ogg_sync_state    sync;
  ogg_page          page;
  ogg_packet        pack;
  char             *buf;
  int               nread, np, sno, nvtracks = 0, natracks = 0, i, n, idx, k;
  int               endofstream = 0;
  struct demux_t    streams[MAX_AUDIO_TRACKS + MAX_VIDEO_TRACKS];
  char             *newName;
  
  memset(streams, 0, sizeof(streams));
  for (i = 0; i < (MAX_AUDIO_TRACKS + MAX_VIDEO_TRACKS); i++)
    streams[i].serial = -1;
  
  ogg_sync_init(&sync);
  while (1) {
    np = ogg_sync_pageseek(&sync, &page);
    if (np < 0) {
      fprintf(stderr, "(%s) ogg_sync_pageseek failed\n", __FILE__);
      return;
    }
    if (np == 0) {
      buf = ogg_sync_buffer(&sync, BLOCK_SIZE);
      if (!buf) {
        fprintf(stderr, "(%s) ogg_sync_buffer failed\n", __FILE__);
        return;
      }
      if ((nread = read(fdin, buf, BLOCK_SIZE)) <= 0) {
        if (verbose > 0)
          fprintf(stdout, "(%s) end of stream\n", __FILE__);
        return;
      }
      ogg_sync_wrote(&sync, nread);
      continue;
    }

    if (!ogg_page_bos(&page)) {
      break;
    } else {
      ogg_stream_state sstate;
      sno = ogg_page_serialno(&page);
      if (ogg_stream_init(&sstate, sno)) {
        fprintf(stderr, "(%s) ogg_stream_init failed\n", __FILE__);
        return;
      }
      ogg_stream_pagein(&sstate, &page);
      ogg_stream_packetout(&sstate, &pack);

      if ((pack.bytes >= 7) && ! strncmp(&pack.packet[1], "vorbis", 6)) {
        if (verbose > 0)
          fprintf(stdout, "(%s) (a%d/%d) found vorbis\n", __FILE__,
                  natracks + 1, nvtracks + natracks);
        if (xaudio[natracks] && (natracks < MAX_AUDIO_TRACKS)) {
          newName = malloc(strlen(basename) + 6);
          if (!newName) {
            fprintf(stderr, "(%s) Failed to allocate %d bytes.\n", __FILE__,
              strlen(basename) + 6);
            exit(1);
          }
          sprintf(newName, "%s-a%d", basename, natracks + 1);
          streams[natracks].fd = open(newName, O_WRONLY | O_CREAT | O_TRUNC,
                                      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
          if (streams[natracks].fd == -1)
            fprintf(stderr, "(%s) Failed to create \"%s\" (%d, %s).\n",
                    __FILE__, newName, errno, strerror(errno));
          else {
            if (verbose > 0)
              fprintf(stdout, "(%s) Extracting a%d to \"%s\".\n", __FILE__,
                      natracks + 1, newName);
            do {
              i = write(streams[natracks].fd, pack.packet, pack.bytes);
              if (verbose > 1)
                fprintf(stdout, "(%s) x/a%d: %d written\n", __FILE__,
                        natracks + 1, i);
            } while (ogg_stream_packetout(&sstate, &pack) == 1);
            streams[natracks].serial = sno;
            streams[natracks].vorbis = 1;
            memcpy(&streams[natracks].state, &sstate, sizeof(sstate));
          }
          free(newName);
        }
        natracks++;
      } else if ((pack.bytes >= 142) &&
                 !strncmp(&pack.packet[1],"Direct Show Samples embedded in Ogg",
                          35) ) {
        if ((*(int32_t*)(pack.packet+96) == 0x05589f80) &&
            (pack.bytes >= 184)) {
          if (verbose > 0)
             fprintf(stdout, "(%s) (v%d/%d) Found old video header. Not " \
                     "supported.\n", __FILE__, nvtracks + 1,
                     nvtracks + natracks + 1);
        } else if (*(int32_t*)pack.packet+96 == 0x05589F81) {
          if (verbose > 0)
            fprintf(stdout, "(%s) (a%d/%d) Found old audio header. Not " \
                    "supported.\n", __FILE__, natracks + 1,
                    nvtracks + natracks + 1);
        } else {
          if (verbose > 0)
            fprintf(stdout, "(%s) OGG stream %d has an old header with an " \
                    "unknown type.", __FILE__, nvtracks + natracks + 1);
        }
      }  else if (((*pack.packet & PACKET_TYPE_BITS ) == PACKET_TYPE_HEADER) &&
	          (pack.bytes >= (int)sizeof(stream_header) + 1)) {
        stream_header *sth = (stream_header *)(pack.packet + 1);
        if (!strncmp(sth->streamtype, "video", 5)) {
          unsigned long codec;
          codec = (sth->subtype[0] << 24) + 
            (sth->subtype[1] << 16) + (sth->subtype[2] << 8) + sth->subtype[3]; 
          if (verbose > 0)
            fprintf(stdout, "(%s) (v%d/%d) fps: %.3f width height: %dx%d " \
                    "codec: %p (%c%c%c%c)\n", __FILE__, nvtracks + 1,
                    natracks + nvtracks + 1,
                    (double)10000000 / (double)sth->time_unit,
                    sth->sh.video.width, sth->sh.video.height, (void *)codec,
                    sth->subtype[0], sth->subtype[1], sth->subtype[2],
                    sth->subtype[3]);
          if (xvideo[nvtracks] && (nvtracks < MAX_VIDEO_TRACKS)) {
            newName = malloc(strlen(basename) + 6);
            idx = MAX_AUDIO_TRACKS + nvtracks;
            if (!newName) {
              fprintf(stderr, "(%s) Failed to allocate %d bytes.\n", __FILE__,
                strlen(basename) + 6);
              exit(1);
            }
            sprintf(newName, "%s-v%d", basename, nvtracks + 1);
            streams[idx].fd = open(newName, O_WRONLY | O_CREAT | O_TRUNC,
                                   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
            if (streams[idx].fd == -1)
              fprintf(stderr, "(%s) Failed to create \"%s\" (%d, %s).\n",
                      __FILE__, newName, errno, strerror(errno));
            else {
              if (verbose > 0)
                fprintf(stdout, "(%s) Extracting v%d to \"%s\".\n", __FILE__,
                        nvtracks + 1, newName);
              do {
                if ((*pack.packet & PACKET_TYPE_HEADER) == 0) {
                  int hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
                  hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
                  i = write(streams[idx].fd, pack.packet + 1 + hdrlen,
                            pack.bytes - 1 - hdrlen);
                  if (verbose > 1)
                    fprintf(stdout, "(%s) x/v%d: %d written\n", __FILE__,
                            nvtracks + 1, i);
                }
              } while (ogg_stream_packetout(&sstate, &pack) == 1);
              streams[idx].serial = sno;
              memcpy(&streams[idx].state, &sstate, sizeof(sstate));
            }
            free(newName);
          }
          nvtracks++;
        } else if (!strncmp(sth->streamtype, "audio", 5)) {
          int codec;
          char buf[5];
          memcpy(buf, sth->subtype, 4);
          buf[4] = 0;
          codec = atoi(buf);
          if (verbose > 0) {
            fprintf(stdout, "(%s) (a%d/%d) codec: %d (0x%04x) (%s), bits per " \
                    "sample: %d channels: %hd  samples per second: %d" \
                    , __FILE__, natracks + 1, natracks + nvtracks + 1,
                    codec, codec,
                    codec == 0x1 ? "PCM" : codec == 55 ? "MP3" :
                    codec == 0x55 ? "MP3" :
                    codec == 0x2000 ? "AC3" : "unknown",
                    sth->bits_per_sample, sth->sh.audio.channels,
                    sth->samples_per_unit);
             fprintf(stdout, " avgbytespersec: %hd blockalign: %d\n",
                     sth->sh.audio.avgbytespersec, sth->sh.audio.blockalign);
          }
          if (xaudio[nvtracks] && (natracks < MAX_AUDIO_TRACKS)) {
            newName = malloc(strlen(basename) + 6);
            idx = natracks;
            if (!newName) {
              fprintf(stderr, "(%s) Failed to allocate %d bytes.\n", __FILE__,
                strlen(basename) + 6);
              exit(1);
            }
            sprintf(newName, "%s-a%d", basename, natracks + 1);
            streams[idx].fd = open(newName, O_WRONLY | O_CREAT | O_TRUNC,
                                   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
            if (streams[idx].fd == -1)
              fprintf(stderr, "(%s) Failed to create \"%s\" (%d, %s).\n",
                      __FILE__, newName, errno, strerror(errno));
            else {
              if (verbose > 0)
                fprintf(stdout, "(%s) Extracting a%d to \"%s\".\n", __FILE__,
                        natracks + 1, newName);
              do {
                if ((*pack.packet & PACKET_TYPE_HEADER) == 0) {
                  int hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
                  hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
                  i = write(streams[idx].fd, pack.packet + 1 + hdrlen,
                            pack.bytes - 1 - hdrlen);
                  if (verbose > 1)
                    fprintf(stdout, "(%s) x/a%d: %d written\n", __FILE__,
                            nvtracks + 1, i);
                }
              } while (ogg_stream_packetout(&sstate, &pack) == 1);
              streams[idx].serial = sno;
              memcpy(&streams[idx].state, &sstate, sizeof(sstate));
            }
            free(newName);
          }
          natracks++;
        } else {
          if (verbose > 0)
            fprintf(stdout, "(%s) (%d) found new header of unknown/" \
                    "unsupported type\n", __FILE__, nvtracks + natracks + 1);
        }

      } else {
        if (verbose > 0)
          fprintf(stdout, "(%s) OGG stream %d is of an unknown type " \
                  "(bad header?)\n", __FILE__, nvtracks + natracks + 1);
      }
    }
  }
  
  endofstream = 0;
  while (!endofstream) {
    sno = ogg_page_serialno(&page);
    for (idx = -1, k = 0; k < (MAX_AUDIO_TRACKS + MAX_VIDEO_TRACKS); k++)
      if (streams[k].serial == sno) {
        idx = k;
        break;
      }
    if (idx == -1) {
      if (verbose > 1)
        fprintf(stdout, "(%s) Encountered packet for an unknown serial " \
                "%d !?\n", __FILE__, sno);
    } else {
      ogg_stream_pagein(&streams[idx].state, &page);
      while (ogg_stream_packetout(&streams[idx].state, &pack) == 1) {
        if (idx < MAX_AUDIO_TRACKS && streams[idx].vorbis)
          n = write(streams[idx].fd, pack.packet, pack.bytes);
        else {
          int hdrlen;
          if (*pack.packet & PACKET_TYPE_HEADER)
            continue;
          hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
          hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
          n = write(streams[idx].fd, pack.packet + 1 + hdrlen,
                    pack.bytes - 1 - hdrlen);
          if (hdrlen && (verbose & 128))
            fprintf(stdout, "(%s) temp - hdrlen is %d\n", __FILE__, hdrlen);
        }

        if (verbose > 1)
          fprintf(stdout, "(%s) x/%c%d: %d written\n", __FILE__,
                  idx >= MAX_AUDIO_TRACKS ? 'v' : 'a',
                  idx >= MAX_AUDIO_TRACKS ? idx - MAX_AUDIO_TRACKS + 1
                                          : idx + 1,
                  n);
      }
    }

    while (ogg_sync_pageseek(&sync, &page) <= 0) {
      buf = ogg_sync_buffer(&sync, BLOCK_SIZE);
      nread = read(fdin, buf, BLOCK_SIZE);
      if (nread <= 0) {
        if (verbose > 0)
          fprintf(stdout, "(%s) end of stream\n", __FILE__);
        endofstream = 1;
        break;
      } else
        ogg_sync_wrote(&sync, nread);
    }
  }
}

int main(int argc, char *argv[]) {

  int i, fdin = -1, aparam = 0, track, vparam = 0;

  memset(xaudio, 1, MAX_AUDIO_TRACKS);
  memset(xvideo, 1, MAX_VIDEO_TRACKS);
  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-i")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -i needs a file name.\n", __FILE__);
        return 1;
      }

      fdin = open(argv[i + 1], O_RDONLY);
      if (fdin == -1) {
        fprintf(stderr, "(%s) Could not open \"%s\".\'", __FILE__,
                argv[i + 1]);
        return 1;
      }
      if (basename == NULL)
        basename = argv[i + 1];
      i++;

    } else if (!strcmp(argv[i], "-v"))
      verbose++;
    else if (!strcmp(argv[i], "-o")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -o needs a file name.\n", __FILE__);
        return 1;
      }
      basename = argv[i + 1];
      i++;
    } else if (!strcmp(argv[i], "-h")) {
      usage(argv[0]);
      return 0;
    } else if (!strcmp(argv[i], "-a")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -a needs a track number.\n", __FILE__);
        return 1;
      }
      if (!aparam) {
        memset(xaudio, 0, MAX_AUDIO_TRACKS);
        aparam = 1;
      }
      track = strtol(argv[i + 1], NULL, 10);
      if ((track < 0) || (track >= MAX_AUDIO_TRACKS)) {
        fprintf(stderr, "(%s) Audio track range is 0..%d.\n", __FILE__,
                MAX_AUDIO_TRACKS);
        return 1;
      }
      xaudio[track] = 1;
      i++;
    } else if (!strcmp(argv[i], "-d")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -d needs a track number.\n", __FILE__);
        return 1;
      }
      if (!vparam) {
        memset(xvideo, 0, MAX_VIDEO_TRACKS);
        vparam = 1;
      }
      track = strtol(argv[i + 1], NULL, 10);
      if ((track < 0) || (track >= MAX_VIDEO_TRACKS)) {
        fprintf(stderr, "(%s) Video track range is 0..%d.\n", __FILE__,
                MAX_VIDEO_TRACKS);
        return 1;
      }
      xvideo[track] = 1;
      i++;
    } else if (!strcmp(argv[i], "-na"))
      memset(xaudio, 0, MAX_AUDIO_TRACKS);
    else if (!strcmp(argv[i], "-nv"))
      memset(xvideo, 0, MAX_VIDEO_TRACKS);
    else {
      fprintf(stderr, "(%s) Unknown parameter \"%s\".\n", __FILE__, argv[i]);
      usage(argv[0]);
    }
  }
  if (fdin == -1) {
    fprintf(stderr, "(%s) No source file specified.\n", __FILE__);
    return 1;
  }
  
  process_ogm(fdin);  
  
  close(fdin);
  return 0;
}
