#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <ogg/ogg.h>
#include <vorbis/codec.h>

#include "ogmstreams.h"

#define MAX_AUDIO_TRACKS 255
#define MAX_VIDEO_TRACKS 255
#define BLOCK_SIZE 4096

struct demux_t {
  int              serial;
  int              fd;
  int              vorbis;
  ogg_stream_state state;
};

int verbose = 0;

void usage(char *fname) {
  fprintf(stdout, "Usage: %s [-hv] inname\n",
          fname);
  fprintf(stdout, "       inname      Use 'inname' as the source.\n");
  fprintf(stdout, "       -v          Be verbose and show each OGG packet.\n");
  fprintf(stdout, "       -h          Show this help.\n");
}

void process_ogm(int fdin)
{
  ogg_sync_state    sync;
  ogg_page          page;
  ogg_packet        pack;
  char             *buf;
  int               nread, np, sno, nvtracks = 0, natracks = 0, i, n, idx, k;
  int               endofstream = 0;
  struct demux_t    streams[MAX_AUDIO_TRACKS + MAX_VIDEO_TRACKS];
  char             *newName;
  
  memset(streams, 0, sizeof(streams));
  for (i = 0; i < (MAX_AUDIO_TRACKS + MAX_VIDEO_TRACKS); i++)
    streams[i].serial = -1;
  
  ogg_sync_init(&sync);
  while (1) {
    np = ogg_sync_pageseek(&sync, &page);
    if (np < 0) {
      fprintf(stderr, "(%s) ogg_sync_pageseek failed\n", __FILE__);
      return;
    }
    if (np == 0) {
      buf = ogg_sync_buffer(&sync, BLOCK_SIZE);
      if (!buf) {
        fprintf(stderr, "(%s) ogg_sync_buffer failed\n", __FILE__);
        return;
      }
      if ((nread = read(fdin, buf, BLOCK_SIZE)) <= 0) {
        if (verbose > 0)
          fprintf(stdout, "(%s) end of stream\n", __FILE__);
        return;
      }
      ogg_sync_wrote(&sync, nread);
      continue;
    }

    if (!ogg_page_bos(&page)) {
      break;
    } else {
      ogg_stream_state sstate;
      sno = ogg_page_serialno(&page);
      if (ogg_stream_init(&sstate, sno)) {
        fprintf(stderr, "(%s) ogg_stream_init failed\n", __FILE__);
        return;
      }
      ogg_stream_pagein(&sstate, &page);
      ogg_stream_packetout(&sstate, &pack);

      if ((pack.bytes >= 7) && ! strncmp(&pack.packet[1], "vorbis", 6)) {
        fprintf(stdout, "(%s) (a%d/%d) Vorbis audio\n", __FILE__,
                natracks + 1, natracks + nvtracks + 1);
        if (verbose > 0)
          do {
            fprintf(stdout, "(%s) a%d: %d bytes\n", __FILE__,
                    natracks, pack.bytes);
          } while (ogg_stream_packetout(&sstate, &pack) == 1);
        streams[natracks].serial = sno;
        streams[natracks].vorbis = 1;
        memcpy(&streams[natracks].state, &sstate, sizeof(sstate));
        natracks++;
      } else if ((pack.bytes >= 142) &&
                 !strncmp(&pack.packet[1],"Direct Show Samples embedded in Ogg",
                          35) ) {
        if ((*(int32_t*)(pack.packet+96) == 0x05589f80) &&
            (pack.bytes >= 184)) {
           fprintf(stdout, "(%s) (v%d/%d) Found old video header. Not " \
                   "supported.\n", __FILE__, nvtracks + 1,
                   natracks + nvtracks + 1);
        } else if (*(int32_t*)pack.packet+96 == 0x05589F81) {
          fprintf(stdout, "(%s) (a%d/%d) Found old audio header. Not " \
                  "supported.\n", __FILE__, natracks + 1,
                  natracks + nvtracks + 1);
        } else {
          if (verbose > 0)
            fprintf(stdout, "(%s) OGG stream %d has an old header with an " \
                    "unknown type.", __FILE__, nvtracks + natracks + 1);
        }
      }  else if (((*pack.packet & PACKET_TYPE_BITS ) == PACKET_TYPE_HEADER) &&
	          (pack.bytes >= (int)sizeof(stream_header) + 1)) {
        stream_header *sth = (stream_header *)(pack.packet + 1);
        if (!strncmp(sth->streamtype, "video", 5)) {
          unsigned long codec;
          codec = (sth->subtype[0] << 24) + 
            (sth->subtype[1] << 16) + (sth->subtype[2] << 8) + sth->subtype[3]; 
          fprintf(stdout, "(%s) (v%d/%d) fps: %.3f width height: %dx%d " \
                  "codec: %p (%c%c%c%c)\n", __FILE__, nvtracks + 1,
                  natracks + nvtracks + 1,
                  (double)10000000 / (double)sth->time_unit,
                  sth->sh.video.width, sth->sh.video.height, (void *)codec,
                  sth->subtype[0], sth->subtype[1], sth->subtype[2],
                  sth->subtype[3]);
          idx = natracks + MAX_AUDIO_TRACKS;
          if (verbose > 0)
            do {
              if ((*pack.packet & PACKET_TYPE_HEADER) == 0) {
                int hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
                hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
                fprintf(stdout, "(%s) v%d: %d bytes\n", __FILE__,
                        nvtracks + 1, pack.bytes - 1 - hdrlen);
              }
            } while (ogg_stream_packetout(&sstate, &pack) == 1);
          streams[idx].serial = sno;
          memcpy(&streams[idx].state, &sstate, sizeof(sstate));
          nvtracks++;
        } else if (!strncmp(sth->streamtype, "audio", 5)) {
          int codec;
          char buf[5];
          memcpy(buf, sth->subtype, 4);
          buf[4] = 0;
          codec = atoi(buf);
          fprintf(stdout, "(%s) (a%d/%d) codec: %d (0x%04x) (%s) bits per " \
                  "sample: %d channels: %hd  samples per second: %d" \
                  , __FILE__, natracks + 1, natracks + nvtracks + 1,
                  codec, codec,
                  codec == 0x1 ? "PCM" : codec == 55 ? "MP3" :
                  codec == 0x55 ? "MP3" :
                  codec == 0x2000 ? "AC3" : "unknown",
                  sth->bits_per_sample, sth->sh.audio.channels,
                  sth->samples_per_unit);
           fprintf(stdout, " avgbytespersec: %hd blockalign: %d\n",
                   sth->sh.audio.avgbytespersec, sth->sh.audio.blockalign);
          idx = natracks;
          if (verbose > 0)
            do {
              if ((*pack.packet & PACKET_TYPE_HEADER) == 0) {
                int hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
                hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
                i = write(streams[idx].fd, pack.packet + 1 + hdrlen,
                          pack.bytes - 1 - hdrlen);
                fprintf(stdout, "(%s) a%d: %d bytes\n", __FILE__,
                        nvtracks + 1, pack.bytes - 1 - hdrlen);
              }
            } while (ogg_stream_packetout(&sstate, &pack) == 1);
          streams[idx].serial = sno;
          memcpy(&streams[idx].state, &sstate, sizeof(sstate));
          natracks++;
        } else {
          fprintf(stdout, "(%s) (%d) found new header of unknown/" \
                  "unsupported type\n", __FILE__, nvtracks + natracks + 1);
        }

      } else {
        fprintf(stdout, "(%s) OGG stream %d is of an unknown type " \
                "(bad header?)\n", __FILE__, nvtracks + natracks + 1);
      }
    }
  }

  if (verbose > 0) {  
    endofstream = 0;
    while (!endofstream) {
      sno = ogg_page_serialno(&page);
      for (idx = -1, k = 0; k < (MAX_AUDIO_TRACKS + MAX_VIDEO_TRACKS); k++)
        if (streams[k].serial == sno) {
          idx = k;
          break;
        }
      if (idx == -1) {
        if (verbose > 1)
          fprintf(stdout, "(%s) Encountered packet for an unknown serial " \
                  "%d !?\n", __FILE__, sno);
      } else {
        ogg_stream_pagein(&streams[idx].state, &page);
        while (ogg_stream_packetout(&streams[idx].state, &pack) == 1) {
          if (idx < MAX_AUDIO_TRACKS && streams[idx].vorbis) {
            n = pack.bytes;
            fprintf(stdout, "(%s) %c%d: %d bytes granulepos: %d\n", __FILE__,
                    idx >= MAX_AUDIO_TRACKS ? 'v' : 'a',
                    idx >= MAX_AUDIO_TRACKS ? idx - MAX_AUDIO_TRACKS + 1
                                            : idx + 1,
                    n, ogg_page_granulepos(&page));
          } else {
            int hdrlen, i;
            long lenbytes = 0;
            hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
            hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
            n = pack.bytes - 1 - hdrlen;
            for (i = 0; i < hdrlen; i++) {
              lenbytes = lenbytes << 8;
              lenbytes += *((unsigned char *)pack.packet + hdrlen - i);
            }
            fprintf(stdout, "(%s) %c%d: %d bytes granulepos: %d ",
                    __FILE__, idx >= MAX_AUDIO_TRACKS ? 'v' : 'a',
                    idx >= MAX_AUDIO_TRACKS ? idx - MAX_AUDIO_TRACKS + 1
                                            : idx + 1,
                    n, ogg_page_granulepos(&page));
            fprintf(stdout, "hdrlen: %d lenbytes: %ld\n", hdrlen, lenbytes);
          }
        }
      }

      while (ogg_sync_pageseek(&sync, &page) <= 0) {
        buf = ogg_sync_buffer(&sync, BLOCK_SIZE);
        nread = read(fdin, buf, BLOCK_SIZE);
        if (nread <= 0) {
          if (verbose > 0)
            fprintf(stdout, "(%s) end of stream\n", __FILE__);
          endofstream = 1;
          break;
        } else
          ogg_sync_wrote(&sync, nread);
      }
    }
  }
}

int main(int argc, char *argv[]) {

  int i, fdin = -1, aparam = 0, track, vparam = 0;

  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-v"))
      verbose++;
    else if (!strcmp(argv[i], "-h")) {
      usage(argv[0]);
      return 0;
    } else {
      fdin = open(argv[i], O_RDONLY);
      if (fdin == -1) {
        fprintf(stderr, "(%s) Could not open \"%s\".\'", __FILE__,
                argv[i]);
        return 1;
      }
    }
  }

  if (fdin == -1) {
    fprintf(stderr, "(%s) No source file specified.\n", __FILE__);
    return 1;
  }
  
  process_ogm(fdin);  
  
  close(fdin);
  return 0;
}
