/*
  ogmmerge -- utility for splicing together ogg bitstreams
      from component media subtypes

  ogmmerge.c

  Rewritten by Moritz Bunkus <moritz@bunkus.org>

  Copyright 2000 Ralph Giles <Ralph_Giles@telus.net>
                 Jack Moffitt <jack@xiph.org>

  Distributed under the GPL
  see http://www.gnu.org/copyleft/gpl.html for details
*/
#define _GNU_SOURCE

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>

#include <ogg/ogg.h>

#include "ogmmerge.h"
#include "util.h"
#include "vorbis.h"
#include "avi.h"
#include "wav.h"

#define VERSIONINFO "ogmmerge v" VERSION "\n"

typedef struct {
  char *ext;
  int   type;
  char *desc;
} file_type_t;

#define TYPENUM 4
file_type_t file_types[TYPENUM] =
  {{"---", TYPEUNKNOWN, "<unknown>"},
   {"ogg", TYPEVORBIS, "Vorbis audio embedded in OGG"},
   {"avi", TYPEAVI, "AVI (Audio/Video Interleaved)"},
   {"wav", TYPEWAV, "WAVE (uncompressed PCM)"}};

param_t params;

static void usage(void)
{
  /* prefer stdout to stderr to coddle win32 */
  FILE *out = stdout;

  fprintf(out,
    VERSIONINFO \
    "ogmmerge [-o out | --output out ] [-tadAD] <file1> [<file2> ...]\n" \
    "\n" \
    " Options for each input file:\n" \
    "  -t, --type <type>       force file type to <type>; see -l.\n" \
    "                          default: auto-detection based on extension.\n" \
    "  -a, --atracks <n,m,...> copy the n'th, m'th... audio track\n" \
    "                          default: copy all audio tracks\n" \
    "  -d, --vtracks <n,m,...> copy the n'th video track, NOT track no. n\n" \
    "                          default: copy all video tracks\n" \
    "  -A, --noaudio           don't copy any audio track from this file\n" \
    "  -D, --novideo           don't copy any video track from this file\n" \
    "\n"
    " Other options:\n"
    "  -l, --list-types        lists supported input file types\n" \
    "  -h, --help              this help\n"
    "  -v, --verbose           verbose status\n"
    "  -q, --quiet             suppress status output\n"
    "  -V, --version           print version information\n"
  );
}

static void set_defaults(void)
{
  /* set defaults */
  params.outfile = NULL;
  params.out = NULL;
  params.input = NULL;
  params.verbose = 1;
}

static int get_type(char *filename)
{
  char *ext;
  int i;

  ext = strrchr(filename, '.');
  if (ext == NULL)
    return TYPEUNKNOWN;
  ext++;

  /* should be smarter and check file magic */
  for (i = 1; i < TYPENUM; i++)
    if (!strcmp(ext, file_types[i].ext))
      return file_types[i].type;
  return TYPEUNKNOWN;
}

static void add_file(filelist_t *file)
{
  filelist_t *temp;

  if (params.input == NULL) {
    params.input = file;
  } else {
    temp = params.input;
    while (temp->next) temp = temp->next;
    temp->next = file;
  }
}

static int unique_serialno(int serialno)
{
  filelist_t *file;

  file = params.input;
  while (file) {
    if (file->serialno == serialno) return 0;
    file = file->next;
  }
  return 1;
}

static char *parse_tracks(char *s) {
  char *c = s;
  char *nstart;
  int n, ntracks;
  unsigned char *tracks;
  
  nstart = NULL;
  tracks = NULL;
  ntracks = 0;
  while (*c) {
    if ((*c >= '0') && (*c <= '9')) {
      if (nstart == NULL)
        nstart = c;
    } else if (*c == ',') {
      *c = 0;
      if (nstart != NULL) {
        n = atoi(nstart);
        if ((n <= 0) || (n > 255)) {
          fprintf(stderr, "Error: track number out of range (1..255): %d\n",
                  n);
          exit(1);
        }
        tracks = (unsigned char *)realloc(tracks, ntracks + 2);
        if (tracks == NULL)
          die("malloc");
        tracks[ntracks] = (unsigned char)n;
        tracks[ntracks + 1] = 0;
        nstart = NULL;
        ntracks++;
      } else
        fprintf(stderr, "Warning: useless use of ','\n");
    } else if (!isspace(*c)) {
      fprintf(stderr, "Error: unrecognized character in track list: '%c'\n",
              *c);
      exit(1);
    }
    c++;
  }
  
  if (nstart != NULL) {
    n = atoi(nstart);
    if ((n <= 0) || (n > 255)) {
      fprintf(stderr, "Error: track number out of range (1..255): %d\n",
              n);
      exit(1);
    }
    tracks = (unsigned char *)realloc(tracks, ntracks + 2);
    if (tracks == NULL)
      die("malloc");
    tracks[ntracks] = (unsigned char)n;
    tracks[ntracks + 1] = 0;
    nstart = NULL;
    ntracks++;
  }
  
  return tracks;
}

static void parse_args(int argc, char **argv)
{
  int         ret;
  int         i, j;
  int         forcetype;
  int         noaudio, novideo;
  char       *atracks, *vtracks;
  filelist_t *file;

  noaudio = 0;
  novideo = 0;
  atracks = NULL;
  vtracks = NULL;
  forcetype = TYPEUNKNOWN;
  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-q"))
      params.verbose = 0;
    else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--verbose"))
      params.verbose = 2;
    else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "-?") ||
             !strcmp(argv[i], "--help")) {
      usage();
      exit(0);
    } else if (!strcmp(argv[i], "-V") || !strcmp(argv[i], "--version")) {
      fprintf(stderr, VERSIONINFO);
      exit(0);
    } else if (!strcmp(argv[i], "-o") || !strcmp(argv[i], "--output")) {
      if ((i + 1) >= argc) {
        fprintf(stderr, "Error: -o lacks a file name.\n");
        exit(1);
      } else if (params.outfile != NULL) {
        fprintf(stderr, "Error: only one output file allowed.\n");
        exit(1);
      }
      params.outfile = (char *)strdup(argv[i + 1]);
      i++;
    } else if (!strcmp(argv[i], "-t") || !strcmp(argv[i], "--type")) {
      if ((i + 1) >= argc) {
        fprintf(stderr, "Error: -t lacks a file type.\n");
        exit(1);
      }
      forcetype = TYPEUNKNOWN;
      for (j = 1; j < TYPENUM; j++)
        if (!strcmp(argv[i + 1], file_types[j].ext)) {
          forcetype = file_types[j].type;
          break;
        }
      if (forcetype == TYPEUNKNOWN) {
        fprintf(stderr, "Error: unknown file type \"%s\" given for -t.\n",
                argv[i + 1]);
        exit(1);
      }
      i++;
    } else if (!strcmp(argv[i], "-l") || !strcmp(argv[i], "--list-types")) {
      fprintf(stdout, "Known file types:\n  ext  description\n" \
                                         "  ---  --------------------------\n");
      for (j = 1; j < TYPENUM; j++)
        fprintf(stdout, "  %s  %s\n", file_types[j].ext, file_types[j].desc);
      exit(0);
    } else if (!strcmp(argv[i], "-A") || !strcmp(argv[i], "--noaudio"))
      noaudio = 1;
    else if (!strcmp(argv[i], "-D") || !strcmp(argv[i], "--novideo"))
      novideo = 1;
    else if (!strcmp(argv[i], "-a") || !strcmp(argv[i], "--atracks")) {
      if ((i + 1) >= argc) {
        fprintf(stderr, "Error: -a lacks the track number(s).\n");
        exit(1);
      }
      if (atracks != NULL)
        free(atracks);
      atracks = parse_tracks(argv[i + 1]);
      i++;
    } else if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--vtracks")) {
      if ((i + 1) >= argc) {
        fprintf(stderr, "Error: -d lacks the track number(s).\n");
        exit(1);
      }
      if (atracks != NULL)
        free(vtracks);
      vtracks = parse_tracks(argv[i + 1]);
      i++;
    } else {
      if ((atracks != NULL) && noaudio) {
        fprintf(stderr, "Error: -A and -a used on the same source file.\n");
        exit(1);
      }
      if ((vtracks != NULL) && novideo) {
        fprintf(stderr, "Error: -D and -d used on the same source file.\n");
        exit(1);
      }
      file = (filelist_t *)mymalloc(sizeof(filelist_t));
      if (file == NULL)
        die("malloc");

      file->name = (char *)strdup(argv[i]);
      if (forcetype != TYPEUNKNOWN)
        file->type = forcetype;
      else
        file->type = get_type(file->name);
      forcetype = TYPEUNKNOWN;

      if (file->type == TYPEUNKNOWN) {
        fprintf(stderr, "Error: File %s has unknown type. Try -t.\n",
                file->name);
        exit(1);
      }

      file->fp = NULL;
      switch (file->type) {
        case TYPEVORBIS:
          file->state_init = vorbis_state_init;
          file->data_in = vorbis_data_in;
          file->page_out = vorbis_page_out;
          break;
        case TYPEAVI:
          file->state_init = avi_state_init;
          file->data_in = avi_data_in;
          file->page_out = avi_page_out;
          break;
        case TYPEWAV:
          file->state_init = wav_state_init;
          file->data_in = wav_data_in;
          file->page_out = wav_page_out;
          break;
      }

      file->serialno = 0;
      file->status = EMOREDATA;
      file->page = NULL;
      file->next = NULL;

      file->atracks = atracks;
      file->vtracks = vtracks;

      add_file(file);
      
      noaudio = 0;
      novideo = 0;
      atracks = NULL;
      vtracks = NULL;
    }
  }
  
  if (params.input == NULL) {
    fprintf(stderr, "Error: no input files given.\n");
    exit(1);
  }
}

#define BUFSIZE 1024

int main(int argc, char **argv)
{
  int i;
  filelist_t *file, *winner;
  unsigned long bytes;
  int serialno;
  int first_pages = 1;
  ogg_page *page;
  char buf[BUFSIZE];

  srand(time(NULL));

  set_defaults();
  parse_args(argc, argv);
  
  /* open output file */
  if (params.outfile == NULL) {
    params.outfile = (char *)strdup("stdout");
    params.out = stdout;
  } else {
    params.out = fopen(params.outfile, "w");
    if (params.out == NULL) {
      fprintf(stderr, "Error: Couldn't open output file %s.\n", params.outfile);
      return 1;
    }
  }

  pq_init();

  /* open all files and prepare for processing */
  file = params.input;
  while (file) {
    fprintf(stderr, "Opening %s for reading (type %s)...\n", file->name,
            file_types[file->type].desc);
    file->fp = fopen(file->name, "r");
    if (file->fp == NULL) {
      fprintf(stderr, "Error: Couldn't open input file %s.\n", file->name);
      return 1;
    }
    
    do {
      serialno = rand();
    } while (!unique_serialno(serialno));
    file->serialno = serialno;
    file->state.name = file->name;
    switch (file->state_init(&file->state, file->serialno, file->atracks,
                             file->vtracks)) {
      case 1:
        file->dontread = 0;
        break;
      case 2:
        file->dontread = 1;
        break;
      default:
        fprintf(stderr, "Filter failed to initialize. Aborting.\n");
        exit(1);
        break;
    }

    file = file->next;
  }

  /* let her rip! */
  while (1) {
    /* Step 1: make sure an ogg page is available for each input 
    ** as long we havne't already processed the last one
    */
    file = params.input;
    while (file) {
      if (file->page == NULL) {
        while ((file->page = file->page_out(&file->state)) == NULL && file->status == EMOREDATA) {
          if (feof(file->fp)) {
            file->status = 0;
            break;
          }
          if (file->dontread)
            file->status = file->data_in(&file->state, NULL, 0);
          else {
            bytes = fread(buf, 1, BUFSIZE, file->fp);
            file->status = file->data_in(&file->state, buf, bytes);
          }
          if (file->status < 0 && file->status != EMOREDATA) {
            fprintf(stderr, "Error: Packetizer error on file %s.\n", file->name);
            return 1;
          }
        }
      }

      file = file->next;
    }

    /* Step 1.5: Write out the first page of each stream
    ** because headers must come together before any
    ** non-header pages.
    */
    if (first_pages) {
      first_pages = 0;
      file = params.input;
      while (file) {
        if (file->page == NULL) {
          fprintf(stderr, "Error: File %s didn't produce a header page.\n", file->name);
          return 1;
        }

        page = file->page->og;

        bytes = fwrite(page->header, 1, page->header_len, params.out);
        if (bytes != page->header_len) {
          fprintf(stderr, "Error: Output error writing to %s.\n", params.outfile);
          return 1;
        }
        bytes = fwrite(page->body, 1, page->body_len, params.out);
        if (bytes != page->body_len) {
          fprintf(stderr, "Error: Output error writing to %s.\n", params.outfile);
          return 1;
        }
      
        myfree(page->header);
        myfree(page->body);
        myfree(page);
        myfree(file->page);
        file->page = NULL;

        file = file->next;
      }

      continue;
    }

    /* Step 2: Pick the page with the lowest timestamp and 
    ** stuff it into the ogg stream
    */
    winner = params.input;
    file = winner->next;
    while (file) {
      if (file->page != NULL) {
        if (winner->page != NULL) {
          if (file->page->timestamp < winner->page->timestamp)
            winner = file;
        } else {
          winner = file;
        }
      }
      file = file->next;
    }

    /* exit if there are no more pages */
    if (winner->page == NULL) break;

    /* Step 3: Write out the winning page */
    page = winner->page->og;

    bytes = fwrite(page->header, 1, page->header_len, params.out);
    if (bytes != page->header_len) {
      fprintf(stderr, "Error: Output error writing to %s.\n", params.outfile);
      return 1;
    }
    bytes = fwrite(page->body, 1, page->body_len, params.out);
    if (bytes != page->body_len) {
      fprintf(stderr, "Error: Output error writing to %s.\n", params.outfile);
      return 1;
    }

    /* Step 4: Cleanup! */
    myfree(page->header);
    myfree(page->body);
    myfree(page);
    myfree(winner->page);
    winner->page = NULL;
  }

  pq_done();

  file = params.input;
  while (file) {
    fclose(file->fp);
    file = file->next;
  }
  fclose(params.out);


  return 0;
}



