/*
  ogmmerge -- utility for splicing together ogg bitstreams
      from component media subtypes

  util.c

  Rewritten by Moritz Bunkus <moritz@bunkus.org>

  Copyright 2000 Ralph Giles <Ralph_Giles@telus.net>
                 Jack Moffitt <jack@xiph.org>

  Distributed under the GPL
  see http://www.gnu.org/copyleft/gpl.html for details
*/

/*
 * mymalloc & myfree for debugging purposes
 */

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>

#include "ogmmerge.h"
#include "util.h"

char *_mymalloc(int size, char *name, int line) {
  char *p = malloc(size);
  fprintf(stderr, "%p MALLOC: %d (%s, %d)\n", p, size, name, line);
  return p;
}

void _myfree(void *p, char *name, int line) {
  fprintf(stderr, "%p FREE (%s, %d)\n", p, name, line);
  free(p);
}

void _die(const char *s, const char *file, int line) {
  fprintf(stderr, "die @ %s/%d : %s\n", file, line, s);
  exit(1);
}

/*
 * Packet queueing stuff
 */

static ogg_page *q_copy_ogg_page(ogg_page *src) {
  ogg_page *dst;

  if (src == NULL)
    die("internal error");
  
  dst = (ogg_page *)mymalloc(sizeof(*dst));
  if (dst == NULL)
    die("malloc");
  dst->header = (unsigned char *)mymalloc(src->header_len);
  if (dst->header == NULL)
    die("malloc");
  dst->body = (unsigned char *)mymalloc(src->body_len);
  if (dst->body == NULL)
    die("malloc");
  dst->header_len = src->header_len;
  dst->body_len = src->body_len;
  memcpy(dst->header, src->header, src->header_len);
  memcpy(dst->body, src->body, src->body_len);
  
  return dst;
}

int q_add_ogg_page(q_t *q, ogg_page *opage) {
  q_page_t *qpage;
  
  if ((q == NULL) || (opage == NULL))
    return EOTHER;
  qpage = (q_page_t *)mymalloc(sizeof(*qpage));
  if (qpage == NULL)
    die("malloc");
  qpage->mpage = (ogmmerge_page_t *)mymalloc(sizeof(*qpage->mpage));
  if (qpage->mpage == NULL)
    die("malloc");
  qpage->mpage->og = q_copy_ogg_page(opage);
  qpage->mpage->timestamp = q->make_timestamp_f(q->private,
                                                ogg_page_granulepos(opage));
  qpage->next = NULL;
  if (q->current != NULL)
    q->current->next = qpage;
  if (q->first == NULL)
    q->first = qpage;
  q->current = qpage;

  return 0;
}

int q_flush_pages(q_t *q) {
  ogg_page *page;
  
  if (q == NULL)
    return EOTHER;
  
  page = (ogg_page *)mymalloc(sizeof(*page));
  
  if (page == NULL)
    die("malloc");
  while (ogg_stream_flush(q->os, page)) {
    q_add_ogg_page(q, page);
    page = (ogg_page *)mymalloc(sizeof(*page));
    if (page == NULL)
      die("malloc");
  }
  myfree(page);
  
  return 0;
}

int q_queue_pages(q_t *q) {
  ogg_page *page;
  
  if (q == NULL)
    return EOTHER;

  page = (ogg_page *)mymalloc(sizeof(*page));
  
  if (page == NULL)
    die("malloc");
  while (ogg_stream_pageout(q->os, page)) {
    q_add_ogg_page(q, page);
    page = (ogg_page *)mymalloc(sizeof(*page));
    if (page == NULL)
      die("malloc");
  }
  myfree(page);
  
  return 0;
}

ogmmerge_page_t *q_page_out(q_t *q)
{
  ogmmerge_page_t *mp;
  struct q_page   *qpage;
  
  if (q == NULL)
    return NULL;

  if (q->first && q->first->mpage) {
    mp = q->first->mpage;
    qpage = q->first->next;
    q->first = qpage;
    return mp;
  }
  
  return NULL;
}

void q_done(q_t *q) {
  q_page_t *qpage, *tmppage;

  if (q == NULL)
    return;  

  qpage = q->first;
  while (qpage) {
    myfree(qpage->mpage->og);
    myfree(qpage->mpage);
    tmppage = qpage->next;
    myfree(qpage);
    qpage = tmppage;
  }
  myfree(q);
}

q_t *q_init(ogmmerge_state_t *oms, ogg_stream_state *os,
            q_make_timestamp_t qtf) {
  q_t *q = (q_t *)mymalloc(sizeof(*q));
  
  if ((oms == NULL) || (os == NULL) || (qtf == NULL))
    return NULL;
  if (q == NULL)
    die("malloc");
  
  q->first = q->current = NULL;
  q->private = oms;
  q->os = os;
  q->make_timestamp_f = qtf;
  
  return q;
}

int q_page_available(q_t *q) {
  if ((q == NULL) || (q->first == NULL) || (q->first->mpage == NULL))
    return 0;
  else
    return 1;
}

/*
 * The big ogmmerge_page queue
 */

static struct q_page *pq_first, *pq_current;

int pq_init() {
  pq_first = NULL;
  pq_current = NULL;
  return 0;
}

int pq_done() {
  /* to be done */
  return 0;
}

int pq_add_ogmmerge_page(ogmmerge_page_t *omp) {
  q_page_t *qpage;
  
  if (omp == NULL)
    return EOTHER;
  qpage = (q_page_t *)mymalloc(sizeof(*qpage));
  if (qpage == NULL)
    die("malloc");
  qpage->mpage = omp;
  qpage->next = NULL;
  if (pq_current != NULL)
   pq_current->next = qpage;
  if (pq_first == NULL)
    pq_first = qpage;
  pq_current = qpage;

  return 0;
}

ogmmerge_page_t *pq_page_out() {
  q_page_t *cur, *last, *winner, *wlast;
  ogmmerge_page_t *omp;
  
  if (pq_first)
    winner = pq_first;
  else
    return NULL;
  last = NULL;
  wlast = NULL;
  cur = pq_first->next;
  while (cur != NULL) {
    if (cur->mpage->timestamp < winner->mpage->timestamp) {
      winner = cur;
      wlast = last;
    }
    last = cur;
    cur = cur->next;
  }
  
  omp = winner->mpage;
  if (wlast != NULL)
    wlast->next = winner->next;
  myfree(winner);
  
  return omp;
}
