/*
  ogmmerge -- utility for splicing together ogg bitstreams
      from component media subtypes

  wav.c
  ogmmerge WAV module

  Written by Moritz Bunkus <moritz@bunkus.org>

  Copyright 2001 Ralph Giles <Ralph_Giles@telus.net>

  Distributed under the GPL
  see the file COPYING for details
  or visit http://www.gnu.org/copyleft/gpl.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <ogg/ogg.h>

#include "ogmmerge.h"
#include "ogmstreams.h"
#include "util.h"
#include "avilib.h" // for wave_header

typedef struct {
  ogg_stream_state    os;
  int                 serialno;
  u_int64_t           packetno;
  u_int64_t           packets_out;
  u_int64_t           old_granulepos;
  int                 bps;
  unsigned long       samples;
  struct wave_header  wheader;
  FILE               *file;  
  unsigned char      *chunk;
  q_t                *q;
} wav_state_t;

#define INTERLEAVE
int interleave = 16;

u_int64_t wav_make_timestamp(ogmmerge_state_t *omst, ogg_int64_t granulepos) {
  wav_state_t *state = omst->private;
  u_int64_t    stamp;

  /* granulepos = seconds * samplerate,
     samplerate = samples_per_unit * channels */
  stamp = granulepos * 1000000 /
          (state->wheader.common.dwSamplesPerSec *
           state->wheader.common.wChannels);
  state->old_granulepos = granulepos;

  return stamp;
}

/* allocates and initializes local storage for a particular
 * substream conversion.
 *
 * returns 1 on success, 0 on failure, 2 on success with the indication
 * that the module handles file access itself
 */
int wav_state_init(ogmmerge_state_t *state, int serialno, char *a, char *v) {
  wav_state_t      *local;
  ogg_packet        op;
  
  if (state == NULL)
    return 0;
  
  local = (wav_state_t *)mymalloc(sizeof(*local));
  if (local == NULL)
    return 0;
  
  ogg_stream_init(&local->os, serialno);
  local->serialno = serialno;
  local->packetno = 0;
  local->packets_out = 0;
  local->old_granulepos = 0;
  state->private = local;
  if ((local->file = fopen(state->name, "r")) == NULL)
    return 0;
  if (fread(&local->wheader, 1, sizeof(local->wheader), local->file) != 
      sizeof(local->wheader))
    return 0;
  if (strncmp(local->wheader.riff.id, "RIFF", 4) ||
      strncmp(local->wheader.riff.wave_id, "WAVE", 4) ||
      strncmp(local->wheader.data.id, "data", 4))
    return 0;
  local->bps = local->wheader.common.wChannels *
               local->wheader.common.wBitsPerSample *
               local->wheader.common.dwSamplesPerSec / 8;
  local->samples = (local->wheader.riff.len - sizeof(local->wheader) + 8) /
                   local->bps;
  local->chunk = mymalloc(local->bps + 1);
  if (local->chunk == NULL)
    die("malloc");
  local->q = q_init(state, &local->os, wav_make_timestamp);
  
  if ((a != NULL) && (v != NULL))
    fprintf(stderr, "Warning: -a/-A/-d/-D have no effect for WAVE files.\n");
  
  return 2;
}

/* closes and deallocates any resources associated
 * with a substream conversion
 *
 * returns 0 on success
 */
int wav_state_close(ogmmerge_state_t *state) {
  wav_state_t *local;
  
  if (state == NULL)
    return 0;
  local = (wav_state_t *)state->private;
  if (local != NULL) {
    ogg_stream_destroy(&local->os);
    state->private = NULL;
    fclose(local->file);
  }
  myfree(local->chunk);
  q_done(local->q);
  myfree(local);

  return 0;
}

/* Normally accepts a buffer of input data 
 * the buffer must be completely consumed before returning
 * 
 * This module handles the file reading itself. buffer will be NULL,
 * size 0.
 *
 * returns 0 on success, a negative error from ogmmerge.h on failure
 */
int wav_data_in(ogmmerge_state_t *state, char *buffer, unsigned long size) {
  wav_state_t          *wavstate;
  ogg_packet           *op;
  ogg_page             *page;
  stream_header         sh;
  struct common_struct *common;
  char                 *buf;
  int                   i;

  if (state == NULL)
    return EOTHER;

  wavstate = (wav_state_t*)state->private;
  
  if (q_page_available(wavstate->q))
    return EMOREDATA;
  
  common = &wavstate->wheader.common;
   
  /* are we looking for the signature? */
  if (wavstate->packetno == 0) {
    memset(&sh, 0, sizeof(sh));
    strcpy(sh.streamtype, "audio");
    memcpy(sh.subtype, "0001", 4);
    sh.size = sizeof(sh);
    sh.time_unit = 10000000;
    sh.samples_per_unit = common->dwSamplesPerSec;
    sh.default_len = 1;
    sh.buffersize = wavstate->bps;
    sh.bits_per_sample = common->wBitsPerSample;
    sh.sh.audio.channels = common->wChannels;
    sh.sh.audio.blockalign = common->wChannels * common->dwSamplesPerSec /
                             interleave;
    sh.sh.audio.avgbytespersec = sh.buffersize;
  
    op = (ogg_packet *)mymalloc(sizeof(*op));
    if (op == NULL)
      die("malloc");
    buf = mymalloc(sizeof(sh) + 1);
    if (buf == NULL)
      die("malloc");
    *buf = PACKET_TYPE_HEADER;
    memcpy(buf + 1, &sh, sizeof(sh));
    op->packet = buf;
    op->bytes = 1 + sh.size;
    op->b_o_s = 1;
    op->e_o_s = 0;
    op->packetno = 0;
    op->granulepos = 0;
    /* submit it */
    ogg_stream_packetin(&wavstate->os, op);
    wavstate->packetno++;
    q_flush_pages(wavstate->q);
  }


  if (wavstate->packetno <= (wavstate->samples * interleave)) {
    int nread, start, j, bytes;
    unsigned short samples;
    unsigned char *bptr;
    nread = fread(wavstate->chunk + 1 + sizeof(samples), 1, wavstate->bps,
                  wavstate->file);
    if (nread < 0)
      return EOTHER;

    for (i = 0; i < interleave; i++) {    
      op = (ogg_packet *)mymalloc(sizeof(*op));
      if (op == NULL)
        die("malloc");
      start = i * (nread / interleave);
      bytes = (i != (interleave - 1) ? nread / interleave : 
              nread - (interleave - 1) * nread / interleave);
      samples = bytes * 8 / common->wBitsPerSample;
      wavstate->chunk[start] = ((sizeof(samples) & 3) << 6) +
                               ((sizeof(samples) & 4) >> 1);
      wavstate->chunk[start] |= (i == 0 ? PACKET_IS_SYNCPOINT : 0);
      op->bytes = bytes + 1 + sizeof(samples);
      bptr = (unsigned char *)&wavstate->chunk[start + 1];
      for (j = 0; j < sizeof(samples); j++) {
        *(bptr + j) = (unsigned char)(samples & 0xFF);
        samples = samples >> 8;
      }
      op->packet = wavstate->chunk + i * (nread / interleave);
      op->b_o_s = 0;
      op->e_o_s = ((wavstate->packetno == wavstate->samples * interleave)
                   && (i == (interleave - 1))) ? 1 : 0;
      /* granulepos = seconds * samplerate,
         samplerate = samples_per_unit * channels */
      op->granulepos = (double)wavstate->packetno / (double)interleave
                       * common->dwSamplesPerSec * common->wChannels;
      op->packetno = wavstate->packetno++;
      ogg_stream_packetin(&wavstate->os, op);
    }
    q_queue_pages(wavstate->q);
    return EMOREDATA;
  }

  return 0;  /* success */
}

/* returns an completed ogmmerge_page created from the data previously
 * passed through mng_data_in()
 *
 * returns NULL if no page is available
 */
ogmmerge_page_t *wav_page_out(ogmmerge_state_t *state) {
  wav_state_t *wavstate = (wav_state_t *)state->private;
  return q_page_out(wavstate->q);
}

